/****************************************************************************
 *
 * Palm Universal Wireless Keyboard(TM) add-on driver for Xwindow. v1.1
 * Copyright GPL Licence. 
 * 2006.01.26 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#define VERSION v1.1
#define MAXNAME 16 
#define PRESSED 1
#define RELEASED 0

#define GRAVE   0x0e
#define ONE     0x16
#define TWO     0x1e
#define THREE   0x26
#define FOUR    0x25
#define FIVE    0x2e
#define SIX     0x36
#define SEVEN   0x3d
#define EIGHT   0x3e
#define NINE    0x46
#define ZERO    0x45
#define MINUS   0x4e
#define EQUAL   0x55
#define BKSP    0x66
#define TAB     0x0d
#define Q       0x15
#define W       0x1d
#define E       0x24
#define R       0x2d
#define T       0x2c
#define Y       0x35
#define U       0x3c
#define I       0x43
#define O       0x44
#define P       0x4d
#define LBRCKT  0x54
#define RBRCKT  0x5b
#define BKSLSH  0x5d
#define CAPSLK  0x58
#define A       0x1c
#define S       0x1b
#define D       0x23
#define F       0x2b
#define G       0x34
#define H       0x33
#define J       0x3b
#define K       0x42
#define L       0x4b
#define SMCOLN  0x4c
#define APSTRP  0x52
#define ENTER   0x5a
#define LSHIFT  0x12
#define Z       0x1a
#define X       0x22
#define C       0x21
#define V       0x2a
#define B       0x32
#define N       0x31
#define M       0x3a
#define COMMA   0x48
#define PERIOD  0x49
#define UP      0x28
#define RSHIFT  0x59
#define SLASH   0x4a
#define LCTRL   0x14
#define LFN     0x02
#define CMD     0x03 
#define CANCEL  0x11
#define SPACE   0x29
#define ALTGR   0x30
#define LEFT    0x5e
#define DOWN    0x60
#define RIGHT   0x2f
#define DEL     0x1f

struct keymapping {
    unsigned char irbyte;

    char nml[MAXNAME];
    unsigned int kcode_nml;   /* keycode normal */

    char lfn[MAXNAME];
    unsigned int kcode_lfn;   /* keycode with LFN */
};  

struct keymapping KeyTable[64] = {
    {GRAVE,         "grave",0,           "",0},
    {ONE,               "1",0,           "",0},
    {TWO,               "2",0,           "",0},
    {THREE,             "3",0,           "",0},
    {FOUR,              "4",0,           "",0},
    {FIVE,              "5",0,           "",0},
    {SIX,               "6",0,           "",0},
    {SEVEN,             "7",0,           "",0},
    {EIGHT,             "8",0,           "",0},
    {NINE,              "9",0,           "",0},
    {ZERO,              "0",0,           "",0},
    {MINUS,         "minus",0,           "",0},
    {EQUAL,         "equal",0,           "",0},
    {BKSP,      "BackSpace",0,           "",0},
    {TAB,             "Tab",0,           "",0},
    {Q,                 "q",0,           "",0},
    {W,                 "w",0,           "",0},
    {E,                 "e",0,           "",0},
    {R,                 "r",0,           "",0},
    {T,                 "t",0,           "",0},
    {Y,                 "y",0,           "",0},
    {U,                 "u",0,           "",0},
    {I,                 "i",0,           "",0},
    {O,                 "o",0,           "",0},
    {P,                 "p",0,           "",0},
    {LBRCKT,  "bracketleft",0,           "",0},
    {RBRCKT, "bracketright",0,           "",0}, 
    {BKSLSH,    "backslash",0,           "",0},
    {CAPSLK,    "Caps_Lock",0,           "",0},
    {A,                 "a",0,           "",0},
    {S,                 "s",0,           "",0},
    {D,                 "d",0,           "",0},
    {F,                 "f",0,           "",0},
    {G,                 "g",0,           "",0},
    {H,                 "h",0,           "",0},
    {J,                 "j",0,           "",0},
    {K,                 "k",0,           "",0},
    {L,                 "l",0,           "",0},
    {SMCOLN,    "semicolon",0,           "",0},
    {APSTRP,   "apostrophe",0,           "",0},
    {ENTER,        "Return",0,           "",0},
    {LSHIFT,      "Shift_L",0,           "",0},
    {Z,                 "z",0,           "",0},
    {X,                 "x",0,           "",0},
    {C,                 "c",0,           "",0},
    {V,                 "v",0,           "",0},
    {B,                 "b",0,           "",0},
    {N,                 "n",0,           "",0},
    {M,                 "m",0,           "",0},
    {COMMA,         "comma",0,           "",0},
    {PERIOD,       "period",0,           "",0},
    {UP,               "Up",0,      "Prior",0},
    {RSHIFT,      "Shift_R",0,           "",0},
    {SLASH,         "slash",0,           "",0},
    {LCTRL,     "Control_L",0,           "",0},
    {CMD,          "Escape",0,           "",0}, 
    {SPACE,         "space",0,           "",0},
    {ALTGR,         "Alt_R",0,           "",0}, 
    {LEFT,           "Left",0,       "Home",0},
    {DOWN,           "Down",0,       "Next",0},
    {RIGHT,         "Right",0,        "End",0},
    {DEL,          "Delete",0,           "",0}, /* 62th */ 
    {CANCEL,             "",0,           "",0},  
    {LFN,                "",0,           "",0}, 
};

int ttyfd;                           /* file descriptor for ttySx*/
struct termios oldtio, newtio;      

unsigned int auto_off_time = 300;     /* seconds */
unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */
unsigned int lfn_status = 0;

Display *d;

void exit_drv(int signum) {

    if ( tcsetattr(ttyfd, TCSAFLUSH, &oldtio) == 0 ) {
	close(ttyfd);
        exit(EXIT_SUCCESS);
    } else { 
	close(ttyfd);
	exit(EXIT_FAILURE);
    }
}

void check_process_dup() {

    FILE *fp;
    char buf[255];

    fp = popen("ps -C uwkxwin -o pid= -o comm= | wc -l", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute ps command.\n");
        exit(EXIT_FAILURE);
    }

    if ( atoi(fgets(buf, 255, fp)) >= 2 ) {
	fprintf(stderr, "uwkxwin is already running!\n");
	exit(EXIT_FAILURE);
    }

    if ( pclose(fp) == -1 ) {
	fprintf(stderr, "unable to close pipe.\n");
	exit(EXIT_FAILURE);
    }
}

void read_keymap() {

    FILE *fp;
    int i;
    char buf[255];
    char cset[] = " \n";
    char *pname, *pcode;
 
    fp = popen("xmodmap -pke", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute xmodmap -pke\n");
        exit(EXIT_FAILURE);
    }

    /* parsing the result of xmodmap -pke */
    while ( fgets(buf, 255, fp) ) {

	strtok(buf, cset);          /* discard "keycode" */
	pcode = strtok(NULL, cset); /* point to X keycode */
	strtok(NULL, cset);         /* discard "="  */
	pname = strtok(NULL, cset); /* point to X keyname */ 


	/* Compare a xmodmap line with UWK 64 keys.        */
	for ( i = 0 ; i <= 63; i++) {  
	    if ( pname != NULL ) {
		if ( !strcmp( KeyTable[i].nml, pname )) {
		    KeyTable[i].kcode_nml = atoi(pcode);
		} 
		if ( !strcmp( KeyTable[i].lfn, pname )) {
		    KeyTable[i].kcode_lfn = atoi(pcode);
		}
	    }
	}
    } 

    if ( pclose(fp) == -1 ) {
	fprintf(stderr, "unable to close pipe.\n");
	exit(EXIT_FAILURE);
    }

    #ifdef DEBUG
    printf("In printing...\n");
    for ( i = 0; i <= 63; i++ ) {
	printf("#%d %x %s %d %s %d\n",
	       i,
	       KeyTable[i].irbyte,
	       KeyTable[i].nml,
	       KeyTable[i].kcode_nml,
	       KeyTable[i].lfn,
	       KeyTable[i].kcode_lfn
	       );
    }
    #endif	       
}
    
void read_rcfile(FILE *fp) {

    char buf[255];    
    while ( (fgets(buf, 255, fp)) != NULL ) {

	if ( buf[0] == '#' ) continue;
	else if ( buf[0] == 'd' ) {
	    sscanf(buf, "%*s %d", &typematic_delay);
	} else if ( buf[0] == 'r' ) {
	    sscanf(buf, "%*s %d", &typematic_rate);
	} else if ( buf[0] == 'a' ) {
	    sscanf(buf, "%*s %d", &auto_off_time);
	} 
    }
    fclose(fp);
}

void translate( unsigned char ch ) {

    int  i;
    unsigned char msb;   /* 0 : key pressed */
                         /* 1 : key released. */

    msb =  (ch & 128) >> 7 ;   
    ch &= 127 ; 

    for (i = 0; i <= 63 ; i++) {

	if ( ch == KeyTable[i].irbyte ) {
	    if ( ch == LFN ) {              
		lfn_status = (1^msb);
		break;
	    } else if ( lfn_status == PRESSED ) {
		if ( ch == DEL ) {          /* LFN-DEL */
		    close(ttyfd);
		    exit(EXIT_SUCCESS);
		}
		if ( ch == EQUAL ) {        /* LFN-EQUAL */
		    if ( msb == 0 ) read_keymap();   /* read when pressed */
		    break;
		}

		XTestFakeKeyEvent(d, KeyTable[i].kcode_lfn, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_lfn);
                #endif


		break;		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].kcode_nml, (1^msb), 0);
		XFlush(d);

                #ifdef DEBUG
		printf("code= %d\n", KeyTable[i].kcode_nml);
                #endif

		break;
	    }
	}
    }
}
	
int main(int argc, char *argv[]) {

    unsigned char ch, irbyte;
    int nread, option;
    FILE *fp;
    char *device = "/dev/ttyS1";
    char *rcfile = "/etc/uwkrc";

    struct sigaction newhandler;
    sigset_t blocked;

    check_process_dup();
    while ( (option = getopt(argc, argv, "hd:c:")) != -1 ) {
        switch(option) {
        case 'h' :
            fprintf(stderr,"Universal Wireless Keyboard(TM) driver v1.1\n"
		           "Usage : uwkxwin [-d device] [-c rcfile]\n"
	                   "Example : uwkxwin -d /dev/ttyS2 -c /etc/uwkrc\n"
		    );
            exit(EXIT_SUCCESS);
        case 'd' :
            device = optarg;
            break;
        case 'c' :
            rcfile = optarg;
            break;
        default :
            fprintf(stderr,"unknown option.\n");
            exit(EXIT_FAILURE);
        }
    }
            

    /* signal install */
    newhandler.sa_handler = exit_drv;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
	perror("sigaction install error\n");
	exit(EXIT_FAILURE);
    }

    /* open irda port device */
    fprintf(stderr, "Opening %s\n", device);
    ttyfd = open(device, O_RDONLY | O_NOCTTY );
    if (ttyfd == -1)  {
        fprintf(stderr, "Cannot open %s\n", device);
        exit(EXIT_FAILURE);
    }

    /* open rcfile */
    fprintf(stderr, "Opening %s\n", rcfile);
    fp = fopen( rcfile, "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot open %s.\n"
                "So, default values are applied.\n", 
		rcfile
		);
    } else {
	read_rcfile( fp );
    }

    /* read current xmodmap */
    read_keymap();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(ttyfd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME] = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]  = 1;   /* blocking read until 1 char received */
    tcsetattr(ttyfd,TCSAFLUSH,&newtio);
               
    d = XOpenDisplay(NULL);
    
    while (1) {
	alarm( auto_off_time );

	nread = read(ttyfd, &ch, 1);
        if ( ( nread > 0 ) && ( ch == 0xff ) ) {
      
	    nread = read(ttyfd, &ch, 1);
	    if ( (nread > 0) && ( ch == 0xc0 ) ) {
	    
		nread = read(ttyfd, &irbyte, 1);
		read(ttyfd, &ch, 1);
		read(ttyfd, &ch, 1); 
		nread = read(ttyfd, &ch, 1);
 
		if ( ch == 0xc1 ) translate(irbyte);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n"
		    "Please test another serial port. Use -h option\n"
		    "Recommend removing other irda modules.\n"
		    );
	    exit(EXIT_FAILURE);
	}         
    }  
}
